
/*@  predicate Swap{L1,L2}(int *a,integer i, integer j)=
  @    \at(a[i],L1) == \at(a[j],L2) &&
  @    \at(a[j],L1) == \at(a[i],L2) &&
  @    (\forall integer k; k!=i && k!=j ==>
  @    \at(a[k],L1) == \at(a[k],L2));
  @
  @  axiomatic permutation {
  @
  @  predicate Permuta{L1,L2} (int *a, integer l, integer h);
  @
  @  axiom Permuta_refl{L}:
  @    \forall int *a, integer l, h; Permuta{L,L} (a,l,h);  
  @  axiom Permuta_sim{L1,L2}:
  @    \forall int *a, integer l, h; 
  @    Permuta{L1,L2} (a,l,h) ==>  Permuta{L2,L1} (a,l,h);
  @  axiom Permuta_trans{L1,L2,L3}:
  @    \forall int *a, integer l, h; 
  @    Permuta{L1,L2} (a,l,h) && Permuta{L2,L3} (a,l,h) ==>
  @    Permuta{L1,L3} (a,l,h);
  @  axiom Permuta_swap{L1,L2}:
  @    \forall int *a, integer l, h, i, j; 
  @    l <= i <= h && l <= j <= h && Swap{L1,L2} (a,i,j) ==>
  @    Permuta{L1,L2} (a,l,h);
  @  }
  @*/



/*@ requires \valid(t+i) && \valid(t+j);
  @ assigns t[i], t[j];
  @ ensures Swap{Old,Here}(t,i,j);
  @*/
void swap(int t[], int i, int j) {
  int tmp = t[i];
  t[i] = t[j];
  t[j] = tmp;
} 

/*@ requires 0 <= p <= r && \valid_range(A,p,r);
  @ assigns A[p..r];
  @ behavior partition:
  @   ensures
  @     p <= \result <= r &&
  @     (\forall int l; p <= l < \result ==>
  @       A[l] <= A[\result]) &&
  @     (\forall int l; \result < l <= r ==>
  @       A[l] > A[\result]) &&
  @     A[\result] == \old(A[r]);
  @ behavior permutation:
  @   ensures
  @     Permuta{Old,Here}(A,p,r) &&
  @     (\forall integer a; (a < p ||  a>r) ==>
  @      A[a] == \at(A[a], Old));
  @*/
int partition (int A[], int p, int r)
{
  int x = A[r];
  int j, i = p-1;
  
  /*@ loop invariant p <= j <= r && p-1 <= i < j;
    @ for partition:
    @   loop invariant
    @     (\forall integer k; (p <= k <= i) ==> A[k] <= x) &&
    @     (\forall integer k; (i <  k <  j) ==> A[k] >  x) &&
    @     A[r] == x;
    @ for permutation:
    @   loop invariant
    @     Permuta{Pre,Here}(A,p,r) && 
    @     (\forall integer u; (u < p || u>=j) ==> 
    @     A[u] == \at(A[u],Pre));
    @ loop variant (r-j);
    @*/
  for(j=p; j<r;j++)
    if(A[j] <= x) {
       i++;
 L1:      swap(A,i,j);
       //@ assert Permuta{L1,Here}(A,p,r);
     }
 L2:  swap(A,i+1,r);
  //@ assert Permuta{L2,Here}(A,p,r);
  return i+1;
}    
